/* ************************************************************** open.c *** *
 * $B2hA|$N%*!<%W%s(B
 *
 * Copyright (C) 2002-2003 Yasuyuki SUGAYA <sugaya@suri.it.okayama-u.ac.jp>
 * Okayama University
 *                                    Time-stamp: <04/06/21 15:21:46 sugaya>
 * ************************************************************************* */
#include "teoeyes.h"
#include "image_io.h"
#include "image_window.h"
#include "open.h"

#include "gtkcustomiconfilesel.h"

/* ************************************************************************* */
gboolean
teoeyes_open_image_on_same_window (const gchar	*filename) {
  GList	*list;

  /* $B%+%l%s%H2hA|$N%a%b%j2rJ|(B */

  /* $B4{$K%j%9%H$K4^$^$l$F$$$k>l9g$O$=$N%G!<%?$rI=<((B */
  list = g_list_first (image_list);
  while (list) {
    if (strcmp (ti_get_file_path (list), filename) == 0) {
      /* $B2hA|$rI=<((B */
      image_list = ti_get_list_from_path (image_list, filename);
      if (!ti_get_pixbuf (image_list)) {
	teoeyes_image_load (ti_get_image (image_list), 0, to->min, to->max);
      }
      gtk_widget_set_size_request (G_GET_WIDGET (image_window, "canvas"),
				   ti_get_width  (image_list),
				   ti_get_height (image_list));
      return TRUE;
    } else {
      list = g_list_next (list);
    }
  }
  /* $B?7$7$$2hA|$r%j%9%H$KDI2C(B */
  if (!teoeyes_image_add_list (g_list_last (image_list), filename, TRUE)) {
    return FALSE;
  }
  /* $B2hA|$rI=<($9$k(B */
  image_list = g_list_last (image_list);
  teoeyes_image_load (ti_get_image (image_list), 0, to->min, to->max);
  gtk_widget_set_size_request (G_GET_WIDGET (image_window, "canvas"),
			       ti_get_width  (image_list),
			       ti_get_height (image_list));
  
  return TRUE;
}

/* $B?7$7$$2hA|$rJL$N%&%#%s%I%&$G3+$/(B **************************************** */
gboolean
teoeyes_open_image_on_new_window (const gchar	*filename) {
  int	pid;

  if (filename) {
    if ((pid = fork ()) == 0) {
      if (execlp (PACKAGE, PACKAGE, filename, (char *) 0) == -1) {
	perror (PACKAGE);
	return FALSE;
      }
    }
  } else {
    if ((pid = fork ()) == 0) {
      if (execlp (PACKAGE, PACKAGE, (char *) 0) == -1) {
	perror (PACKAGE);
	return FALSE;
      }
    }
  }
  return TRUE;
}

/* OK$B%\%?%s$,2!$5$l$?;~$N=hM}(B ********************************************** */
static void
cb_dialog_ok (GtkWidget			*widget,
	      GtkCustomIconFileSel	*dialog) {
  gboolean	flag;
  gchar		*filename, *real_path;
  
  /* $B?7$7$$%&%#%s%I%&$r3+$/$+$I$&$+D4$Y$k(B */
  flag =
    gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON 
				  (gtk_object_get_data (GTK_OBJECT (dialog),
							"checkbutton")));
  /* $B%U%!%$%kL>$r3MF@$9$k(B */
  filename = (gchar *) 
    gtk_custom_icon_file_selection_get_filename (GTK_CUSTOM_ICON_FILESEL
						 (dialog));
  real_path = g_filename_from_utf8 (filename, -1, NULL, NULL, NULL);
  if (!real_path || strcmp (real_path, "") == 0 ||
      !g_file_exist (real_path) || g_dir_is (real_path)) return;  

  /* $B;XDj$5$l$?%U%!%$%k$,BP1~$7$?2hA|%U%!%$%k$+$I$&$+D4$Y$k(B */
  if (!check_image_supported (real_path)) return;

  /* $B2hA|$r3+$/(B */
  if (flag) {
    teoeyes_open_image_on_new_window (real_path);
  } else {
    teoeyes_open_image_on_same_window (real_path);
  }
  g_free (real_path);
}

/* Cancel$B%\%?%s$,2!$5$l$?;~$N=hM}(B ****************************************** */
static void 
cb_dialog_cancel (GtkWidget	*widget,
		  GtkWidget	*dialog) {
  gtk_widget_hide (dialog);
}

/* ************************************************************************* */
static void
cb_select_item (GtkFileList		*flist,
		GtkFileListItem		*icon,
		GdkEventButton		*ev,
		gpointer		data) {
  if (ev->type == GDK_2BUTTON_PRESS) {
    cb_dialog_ok (NULL, GTK_CUSTOM_ICON_FILESEL (data));
  }
}

/* ************************************************************************* */
static void
open_dialog_destroy (GtkWidget	*widget,
		     gpointer	data) { 
  gtk_widget_hide (widget);
}

/* $B2hA|%*!<%W%s%@%$%"%m%0$N@8@.(B ******************************************** */
GtkWidget*
open_dialog_new (const gchar	*title) {
  GtkWidget	*dialog;
  GtkWidget	*widget;

  /* $B%@%$%"%m%0$N@8@.(B */
  dialog = gtk_custom_icon_file_selection_new (title);
  gtk_window_set_icon (GTK_WINDOW (dialog), te_icon);
  g_signal_connect (G_OBJECT (dialog), "destroy",
		    G_CALLBACK (open_dialog_destroy), NULL);
  g_signal_connect (G_OBJECT (dialog), "delete_event",
		    G_CALLBACK (open_dialog_destroy), NULL);
  
  /* $B%A%'%C%/%\%?%s$NDI2C(B */
  widget = gtk_check_button_new_with_label (_("Open on a new window"));
  G_SET_WIDGET (dialog, "checkbutton", widget);
  gtk_box_pack_start (GTK_BOX
		      (GTK_CUSTOM_ICON_FILESEL (dialog)->custom_box[6]),
		      widget, FALSE, FALSE, 0);
  gtk_widget_show (GTK_CUSTOM_ICON_FILESEL (dialog)->custom_box[6]);

  /* $B%$%Y%s%H$N@_Dj(B */
  g_signal_connect (G_OBJECT (GTK_CUSTOM_ICON_FILESEL (dialog)->ok_button),
		    "clicked", G_CALLBACK (cb_dialog_ok), dialog);
  g_signal_connect (G_OBJECT (GTK_CUSTOM_ICON_FILESEL (dialog)->cancel_button),
		    "clicked", G_CALLBACK (cb_dialog_cancel), dialog);
  g_signal_connect (G_OBJECT (GTK_CUSTOM_ICON_FILESEL (dialog)->file_list),
		    "select_icon", G_CALLBACK (cb_select_item), dialog);
  
  return dialog;
}

/* ******************************************************* End of open.c *** */
